###############################################################################
#
# Copyright (c) Crossbar.io Technologies GmbH and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at https://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
###############################################################################

import uuid
import binascii

import cbor2
import nacl.secret
import nacl.utils
import nacl.exceptions
import nacl.public

import txaio
from autobahn.twisted.util import sleep

import eth_keys
from eth_account import Account

from ._interfaces import IConsumer, IBuyer


class SimpleBuyer(object):
    """
    Simple XBR buyer component. This component can be used by a XBR consumer to handle
    XBR buying transactions to buy data keys for services used by the XBR consumer.
    """

    log = txaio.make_logger()

    def __init__(self, buyer_key):
        """

        :param buyer_key: Consumer delegate (buyer) private Ethereum key.
        :type buyer_key: bytes
        """
        self._pkey = eth_keys.keys.PrivateKey(buyer_key)
        self._acct = Account.privateKeyToAccount(self._pkey)

        # this holds the keys we bought (map: key_id => nacl.secret.SecretBox)
        self._keys = {}
        self._session = None
        self._running = False

        self._receive_key = nacl.public.PrivateKey.generate()

    async def start(self, session, consumer_id):
        """

        :param session:
        :param consumer_id:
        :return:
        """
        self._session = session
        self._running = True

        self.log.info('Start buying from consumer delegate address {address} (public key 0x{public_key}..)',
                      address=self._acct.address,
                      public_key=binascii.b2a_hex(self._pkey.public_key[:10]).decode())

        channels = await session.call('xbr.marketmaker.get_payment_channels', self._acct.address)
        if not channels:
            raise Exception('no active payment channels found')

        channel_info = await session.call('xbr.marketmaker.get_payment_channel', channels[0])

        if not channel_info or not channel_info.get('id', None):
            raise Exception('no payment channel found')
        if channel_info['status'] != 'open':
            raise Exception('payment channel not open')
        if not channel_info.get('balance', None):
            raise Exception('no positive balance in payment channel')

        self._channel = channel_info['id']
        self._channel_seq = channel_info['seq']
        self._balance = channel_info['balance']

        return self._balance

    async def unwrap(self, key_id, enc_ser, ciphertext):
        """

        :param key_id:
        :param enc_ser:
        :param ciphertext:
        :return:
        """
        assert(enc_ser == 'cbor')

        # if we don't have the key, buy it!
        if key_id not in self._keys:
            # mark the key as currently being bought already (the location of code here is multi-entrant)
            self._keys[key_id] = False

            # call the market maker to buy the key
            amount = 35
            balance = 0

            self._channel_seq += 1

            # FIXME: compute actual kecchak256 based signature
            signature = b'\x00' * 64

            buyer_pubkey = self._receive_key.public_key.encode(encoder=nacl.encoding.RawEncoder)

            # call the market maker to buy the key
            #   -> channel_id, channel_seq, buyer_pubkey, datakey_id, amount, balance, signature
            sealed_key = await self._session.call('xbr.marketmaker.buy',
                                                  self._channel,
                                                  self._channel_seq,
                                                  buyer_pubkey,
                                                  key_id,
                                                  amount,
                                                  balance,
                                                  signature)

            unseal_box = nacl.public.SealedBox(self._receive_key)
            try:
                key = unseal_box.decrypt(sealed_key)
            except nacl.exceptions.CryptoError as e:
                raise Exception('failed to unseal the datakey: {}'.format(e))

            # remember the key, so we can use it to actually decrypt application payload data
            self._keys[key_id] = nacl.secret.SecretBox(key)
            self.log.info('Key {key_id} bought!', key_id=uuid.UUID(bytes=key_id))

        # if the key is already bein bought, wait until the one buying string of execution has succeeded and done
        while self._keys[key_id] == False:
            self.log.info('Waiting for key {key_id} currently being bought ..', key_id=uuid.UUID(bytes=key_id))
            await sleep(.2)

        # now that we have the secret key, decrypt the event application payload
        try:
            message = self._keys[key_id].decrypt(ciphertext)
        except nacl.exceptions.CryptoError as e:
            # Decryption failed. Ciphertext failed verification
            raise RuntimeError('unwrapping XBR payload failed ({})'.format(e))

        try:
            payload = cbor2.loads(message)
        except cbor2.decoder.CBORDecodeError as e:
            # premature end of stream (expected to read 4187 bytes, got 27 instead)
            raise RuntimeError('unwrapping XBR payload failed ({})'.format(e))

        return payload


IBuyer.register(SimpleBuyer)
IConsumer.register(SimpleBuyer)



###############################################################################
#
# Copyright (c) Crossbar.io Technologies GmbH and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at https://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
###############################################################################

from __future__ import absolute_import

from xbr._version import __version__
from xbr._abi import XBR_TOKEN_ABI, XBR_NETWORK_ABI, XBR_PAYMENT_CHANNEL_ABI
from xbr._abi import XBR_DEBUG_TOKEN_ADDR, XBR_DEBUG_NETWORK_ADDR
from xbr._buyer import SimpleBuyer
from xbr._seller import SimpleSeller
from xbr._interfaces import IMarketMaker, IProvider, IConsumer, ISeller, IBuyer


version = __version__
"""
XBR library version.
"""

xbrToken = None
"""
Contract instance of the token used in the XBR Network.
"""

xbrNetwork = None
"""
Contract instance of the XBR Network.
"""


def setProvider(_w3):
    """
    The XBR library must be initialized (once) first by setting the Web3 provider
    using this function.
    """
    global xbrToken
    global xbrNetwork
    xbrToken = _w3.eth.contract(address=XBR_DEBUG_TOKEN_ADDR, abi=XBR_TOKEN_ABI)
    xbrNetwork = _w3.eth.contract(address=XBR_DEBUG_NETWORK_ADDR, abi=XBR_NETWORK_ABI)


class MemberLevel(object):
    """
    XBR Network member levels.
    """
    NONE = 0
    ACTIVE = 1
    VERIFIED = 2
    RETIRED = 3
    PENALTY = 4
    BLOCKED = 5


class NodeType(object):
    """
    XBR Cloud node types.
    """
    NONE = 0
    MASTER = 1
    CORE = 2
    EDGE = 3


class ActorType(object):
    """
    XBR Market actor types.
    """
    NONE = 0
    NETWORK = 1
    MARKET = 2
    PROVIDER = 3
    CONSUMER = 4


__all__ = (
    'version',
    'setProvider',

    'xbrToken',
    'xbrNetwork',

    'MemberLevel',
    'ActorType',
    'NodeType',

    'XBR_TOKEN_ABI',
    'XBR_NETWORK_ABI',
    'XBR_PAYMENT_CHANNEL_ABI',

    'IMarketMaker',
    'IProvider',
    'IConsumer',
    'ISeller',
    'IBuyer',

    'SimpleBuyer',
    'SimpleSeller',
)



###############################################################################
#
# Copyright (c) Crossbar.io Technologies GmbH and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at https://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
###############################################################################

import os
import time
import uuid
import binascii

import cbor2
import nacl.secret
import nacl.utils

from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall

import txaio

import zlmdb

from autobahn.wamp.types import RegisterOptions
from autobahn.wamp.exception import ApplicationError, TransportLost
from autobahn.twisted.util import sleep


import eth_keys
from eth_account import Account

from ._interfaces import IProvider, ISeller

import click


def hl(text, bold=True, color='yellow'):
    if not isinstance(text, str):
        text = '{}'.format(text)
    return click.style(text, fg=color, bold=bold)


class SimpleSeller(object):

    log = txaio.make_logger()

    def __init__(self, private_key, interval, price):
        """

        :param private_key:
        :param interval:
        :param price:
        """
        self._pkey = eth_keys.keys.PrivateKey(private_key)
        self._acct = Account.privateKeyToAccount(self._pkey)
        self._interval = interval
        self._price = price
        self._id = None
        self._key = None
        self._box = None
        self._archive = {}
        self._run_loop = None
        self._started = None
        self._session = None

    @property
    def public_key(self):
        return self._pkey.public_key

    def start(self, session, provider_id):
        """

        :param session:
        :param provider_id:
        :return:
        """
        assert self._run_loop is None
        assert self._session is None

        self._session = session

        dl = []
        for func in [self.sell]:
            procedure = 'xbr.provider.{}.{}'.format(provider_id, func.__name__)
            d = session.register(func, procedure, options=RegisterOptions(details_arg='details'))
            dl.append(d)

        d = txaio.gather(dl)

        def registered(regs):
            for reg in regs:
                self.log.info('Registered procedure "{procedure}"', procedure=hl(reg.procedure))

        d.addCallback(registered)

        self._run_loop = LoopingCall(self.loop, session)

        self._started = self._run_loop.start(self._interval)

        def on_stop(res_or_err):
            self._run_loop = None
            # self._started = None
            self.log.info('Stopped selling from provider delegate address "{address}" (public key "0x{public_key}..")',
                          address=hl(self._acct.address),
                          public_key=hl(binascii.b2a_hex(self._pkey.public_key[:10]).decode()))

        self._started.addBoth(on_stop)

        self.log.info('Started selling from provider delegate address "{address}" (public key "0x{public_key}..")',
                      address=hl(self._acct.address),
                      public_key=hl(binascii.b2a_hex(self._pkey.public_key[:10]).decode()))

        return self._started

    def stop(self):
        """

        :return:
        """
        if self._run_loop:
            self._run_loop.stop()

        if self._session and self._session.is_attached():
            # FIXME: voluntarily unregister interface
            pass

        return self._started

    async def wrap(self, uri, payload):
        """

        :param uri:
        :param payload:
        :return:
        """
        data = cbor2.dumps(payload)
        ciphertext = self._box.encrypt(data)
        return self._id, 'cbor', ciphertext

    # def unwrap(self, key_id, enc_ser, ciphertext):
    #     assert(enc_ser == 'cbor')
    #     data = self.decrypt(key_id, ciphertext)
    #     return cbor2.loads(data)
    #
    # def encrypt(self, data):
    #     assert(type(data) == bytes)
    #     assert(self._box is not None)
    #     return self._id, self._box.encrypt(data)
    #
    # def decrypt(self, key_id, data):
    #     assert(type(data) == bytes)
    #     assert(key_id in self._archive)
    #     return self._archive[key_id][2].decrypt(data)

    def _rotate(self):
        self._id = os.urandom(16)
        self._key = nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)
        self._box = nacl.secret.SecretBox(self._key)
        self._archive[self._id] = (zlmdb.time_ns(), self._key, self._box)
        self.log.info('key rotated, new key_id={key_id}', key_id=uuid.UUID(bytes=self._id))

    @inlineCallbacks
    def loop(self, session):
        # rotate our key
        self._rotate()

        # offer the key to the market maker (retry 5x in specific error cases)
        retries = 5
        while retries:
            try:
                # FIXME
                api_id = b'\0' * 16
                uri = 'debug.'
                valid_from = time.time_ns() - 10 * 10**9
                signature = None

                offer = yield session.call('xbr.marketmaker.place_offer',
                                           self._id,
                                           api_id,
                                           uri,
                                           valid_from,
                                           signature,
                                           price=self._price)
                self.log.info('New key {key_id} offered at XBR market maker: {offer}', key_id=None, offer=offer)
                break

            except ApplicationError as e:
                if e.error == 'wamp.error.no_such_procedure':
                    self.log.warn('xbr.marketmaker.offer: procedure unavailable!')
                else:
                    self.log.failure()
                    break
            except TransportLost:
                self.log.warn('TransportLost while calling xbr.marketmaker.offer!')
                break
            except:
                self.log.failure()

            retries -= 1
            self.log.warn('Failed to place offer for key! Retrying {retries}/5 ..', retries=retries)
            yield sleep(1)

    def sell(self, key_id, buyer_pubkey, amount_paid, post_balance, signature, details=None):
        """

        :param key_id:
        :param buyer_pubkey:
        :param amount_paid:
        :param post_balance:
        :param signature:
        :param details:
        :return:
        """
        if key_id not in self._archive:
            raise RuntimeError('no such datakey')

        created, key, box = self._archive[key_id]

        # FIXME: check amount paid, post balance and signature
        # FIXME: sign transaction

        sendkey_box = nacl.public.SealedBox(nacl.public.PublicKey(buyer_pubkey,
                                                                  encoder=nacl.encoding.RawEncoder))

        encrypted_key = sendkey_box.encrypt(key, encoder=nacl.encoding.RawEncoder)

        self.log.info('Key {key_id} sold to {buyer_pubkey} (caller={caller})', key_id=key_id, caller=details.caller, buyer_pubkey=buyer_pubkey)

        return encrypted_key


ISeller.register(SimpleSeller)
IProvider.register(SimpleSeller)



###############################################################################
#
# Copyright (c) Crossbar.io Technologies GmbH and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
###############################################################################

from __future__ import absolute_import
from setuptools import setup

# read version string
__version__ = None
with open('xbr/_version.py') as f:
    exec(f.read())  # defines __version__

# read package long description
with open('README.rst') as f:
    docstr = f.read()

setup(
    name='xbr',
    version=__version__,
    description='The XBR Protocol - blockchain protocol for decentralized open data markets',
    long_description=docstr,
    license='Apache 2.0 License',
    author='Crossbar.io Technologies GmbH',
    author_email='support@crossbario.com',
    url='https://xbr.network',
    platforms='Any',
    install_requires=[
        'cbor2>=4.1.1',         # MIT license
        'zlmdb>=18.12.1',       # MIT license
        'twisted>=18.9.0',      # MIT license
        'autobahn>=18.11.2',    # MIT license
        'web3>=4.8.1',          # MIT license
    ],
    extras_require={},
    packages=[
        'xbr',
    ],
    package_data={'xbr': ['./xbr/contracts/*.json']},

    # this flag will make files from MANIFEST.in go into _source_ distributions only
    include_package_data=True,
    zip_safe=False,

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=["License :: OSI Approved :: Apache Software License",
                 "Development Status :: 3 - Alpha",
                 "Environment :: No Input/Output (Daemon)",
                 "Intended Audience :: Developers",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 2",
                 "Programming Language :: Python :: 2.7",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.4",
                 "Programming Language :: Python :: 3.5",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: Implementation :: CPython",
                 "Programming Language :: Python :: Implementation :: PyPy",
                 "Programming Language :: Python :: Implementation :: Jython",
                 "Topic :: Internet",
                 "Topic :: Internet :: WWW/HTTP",
                 "Topic :: Communications",
                 "Topic :: System :: Distributed Computing",
                 "Topic :: Software Development :: Libraries",
                 "Topic :: Software Development :: Libraries :: Python Modules",
                 "Topic :: Software Development :: Object Brokering"],
    keywords='xbr data-markets blockchain ethereum wamp autobahn crossbar'
)
